
#ifndef _FEATURES_TO_RECONF_H_
#define _FEATURES_TO_RECONF_H_

#include "features.h"

namespace CMSat {

int get_reconf_from_features(const Features& feat, const int verb);

} //end namespace

#endif //_FEATURES_TO_RECONF_H_


